/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import { __assign, __awaiter, __generator } from "tslib";
import { QuadSchema, QuadValues } from '../../../mol-gl/compute/util';
import { createComputeRenderable } from '../../../mol-gl/renderable';
import { DefineSpec, TextureSpec, UniformSpec } from '../../../mol-gl/renderable/schema';
import { ShaderCode } from '../../../mol-gl/shader-code';
import quad_vert from '../../../mol-gl/shader/quad.vert';
import { createComputeRenderItem } from '../../../mol-gl/webgl/render-item';
import { ValueCell } from '../../../mol-util';
import { arrayMin } from '../../../mol-util/array';
import { isLittleEndian } from '../../../mol-util/is-little-endian';
import { normalizeBasicOrder } from '../spherical-functions';
import shader_frag from './shader.frag';
var AlphaOrbitalsSchema = __assign(__assign({}, QuadSchema), { uDimensions: UniformSpec('v3'), uMin: UniformSpec('v3'), uDelta: UniformSpec('v3'), tCenters: TextureSpec('image-float32', 'rgba', 'float', 'nearest'), tInfo: TextureSpec('image-float32', 'rgba', 'float', 'nearest'), tCoeff: TextureSpec('image-float32', 'rgb', 'float', 'nearest'), tAlpha: TextureSpec('image-float32', 'alpha', 'float', 'nearest'), uWidth: UniformSpec('f'), uNCenters: UniformSpec('i'), uNAlpha: UniformSpec('i'), uNCoeff: UniformSpec('i'), uMaxCoeffs: UniformSpec('i'), uLittleEndian: UniformSpec('b'), uDensity: UniformSpec('b'), uOccupancy: UniformSpec('f'), tCumulativeSum: TextureSpec('texture', 'rgba', 'ubyte', 'nearest') });
var AlphaOrbitalsName = 'alpha-orbitals';
var AlphaOrbitalsTex0 = 'alpha-orbitals-0';
var AlphaOrbitalsTex1 = 'alpha-orbitals-1';
var AlphaOrbitalsShaderCode = ShaderCode(AlphaOrbitalsName, quad_vert, shader_frag);
function getNormalizedAlpha(basis, alphaOrbitals, sphericalOrder) {
    var alpha = new Float32Array(alphaOrbitals.length);
    var aO = 0;
    for (var _i = 0, _a = basis.atoms; _i < _a.length; _i++) {
        var atom = _a[_i];
        for (var _b = 0, _c = atom.shells; _b < _c.length; _b++) {
            var shell = _c[_b];
            for (var _d = 0, _e = shell.angularMomentum; _d < _e.length; _d++) {
                var L = _e[_d];
                var a0 = normalizeBasicOrder(L, alphaOrbitals.slice(aO, aO + 2 * L + 1), sphericalOrder);
                for (var i = 0; i < a0.length; i++)
                    alpha[aO + i] = a0[i];
                aO += 2 * L + 1;
            }
        }
    }
    return alpha;
}
function createTextureData(grid, orbital) {
    var _a = grid.params, basis = _a.basis, sphericalOrder = _a.sphericalOrder, cutoffThreshold = _a.cutoffThreshold;
    var centerCount = 0;
    var baseCount = 0;
    var coeffCount = 0;
    for (var _i = 0, _b = basis.atoms; _i < _b.length; _i++) {
        var atom = _b[_i];
        for (var _c = 0, _d = atom.shells; _c < _d.length; _c++) {
            var shell = _d[_c];
            for (var _e = 0, _f = shell.angularMomentum; _e < _f.length; _e++) {
                var L = _f[_e];
                if (L > 4) {
                    // TODO: will L > 4 be required? Would need to precompute more functions in that case.
                    throw new Error('Angular momentum L > 4 not supported.');
                }
                centerCount++;
                baseCount += 2 * L + 1;
                coeffCount += shell.exponents.length;
            }
        }
    }
    var centers = new Float32Array(4 * centerCount);
    // L, alpha_offset, coeff_offset_start, coeff_offset_end
    var info = new Float32Array(4 * centerCount);
    var alpha = new Float32Array(baseCount);
    var coeff = new Float32Array(3 * coeffCount);
    var maxCoeffs = 0;
    var cO = 0, aO = 0, coeffO = 0;
    for (var _g = 0, _h = basis.atoms; _g < _h.length; _g++) {
        var atom = _h[_g];
        for (var _j = 0, _k = atom.shells; _j < _k.length; _j++) {
            var shell = _k[_j];
            var amIndex = 0;
            for (var _l = 0, _m = shell.angularMomentum; _l < _m.length; _l++) {
                var L = _m[_l];
                var a0 = normalizeBasicOrder(L, orbital.alpha.slice(aO, aO + 2 * L + 1), sphericalOrder);
                var cutoffRadius = cutoffThreshold > 0
                    ? Math.sqrt(-Math.log(cutoffThreshold) / arrayMin(shell.exponents))
                    : 10000;
                centers[4 * cO + 0] = atom.center[0];
                centers[4 * cO + 1] = atom.center[1];
                centers[4 * cO + 2] = atom.center[2];
                centers[4 * cO + 3] = cutoffRadius * cutoffRadius;
                info[4 * cO + 0] = L;
                info[4 * cO + 1] = aO;
                info[4 * cO + 2] = coeffO;
                info[4 * cO + 3] = coeffO + shell.exponents.length;
                for (var i = 0; i < a0.length; i++)
                    alpha[aO + i] = a0[i];
                var c0 = shell.coefficients[amIndex++];
                for (var i = 0; i < shell.exponents.length; i++) {
                    coeff[3 * (coeffO + i) + 0] = c0[i];
                    coeff[3 * (coeffO + i) + 1] = shell.exponents[i];
                }
                if (c0.length > maxCoeffs) {
                    maxCoeffs = c0.length;
                }
                cO++;
                aO += 2 * L + 1;
                coeffO += shell.exponents.length;
            }
        }
    }
    return { nCenters: centerCount, nAlpha: baseCount, nCoeff: coeffCount, maxCoeffs: maxCoeffs, centers: centers, info: info, alpha: alpha, coeff: coeff };
}
function createAlphaOrbitalsRenderable(ctx, grid, orbital) {
    var data = createTextureData(grid, orbital);
    var _a = grid.dimensions, nx = _a[0], ny = _a[1], nz = _a[2];
    var width = Math.ceil(Math.sqrt(nx * ny * nz));
    if (!ctx.namedFramebuffers[AlphaOrbitalsName]) {
        ctx.namedFramebuffers[AlphaOrbitalsName] = ctx.resources.framebuffer();
    }
    if (!ctx.namedTextures[AlphaOrbitalsTex0]) {
        ctx.namedTextures[AlphaOrbitalsTex0] = ctx.resources.texture('image-uint8', 'rgba', 'ubyte', 'nearest');
    }
    if (!ctx.namedTextures[AlphaOrbitalsTex1]) {
        ctx.namedTextures[AlphaOrbitalsTex1] = ctx.resources.texture('image-uint8', 'rgba', 'ubyte', 'nearest');
    }
    var values = __assign(__assign({}, QuadValues), { uDimensions: ValueCell.create(grid.dimensions), uMin: ValueCell.create(grid.box.min), uDelta: ValueCell.create(grid.delta), uWidth: ValueCell.create(width), uNCenters: ValueCell.create(data.nCenters), uNAlpha: ValueCell.create(data.nAlpha), uNCoeff: ValueCell.create(data.nCoeff), uMaxCoeffs: ValueCell.create(data.maxCoeffs), tCenters: ValueCell.create({ width: data.nCenters, height: 1, array: data.centers }), tInfo: ValueCell.create({ width: data.nCenters, height: 1, array: data.info }), tCoeff: ValueCell.create({ width: data.nCoeff, height: 1, array: data.coeff }), tAlpha: ValueCell.create({ width: data.nAlpha, height: 1, array: data.alpha }), uLittleEndian: ValueCell.create(isLittleEndian()), uDensity: ValueCell.create(false), uOccupancy: ValueCell.create(0), tCumulativeSum: ValueCell.create(ctx.namedTextures[AlphaOrbitalsTex1]) });
    var schema = __assign({}, AlphaOrbitalsSchema);
    if (!ctx.isWebGL2) {
        // workaround for webgl1 limitation that loop counters need to be `const`
        schema.uNCenters = DefineSpec('number');
        schema.uMaxCoeffs = DefineSpec('number');
    }
    var renderItem = createComputeRenderItem(ctx, 'triangles', AlphaOrbitalsShaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
function getAlphaOrbitalsRenderable(ctx, grid, orbital) {
    if (ctx.namedComputeRenderables[AlphaOrbitalsName]) {
        var v = ctx.namedComputeRenderables[AlphaOrbitalsName].values;
        var data = createTextureData(grid, orbital);
        var _a = grid.dimensions, nx = _a[0], ny = _a[1], nz = _a[2];
        var width = Math.ceil(Math.sqrt(nx * ny * nz));
        ValueCell.update(v.uDimensions, grid.dimensions);
        ValueCell.update(v.uMin, grid.box.min);
        ValueCell.update(v.uDelta, grid.delta);
        ValueCell.updateIfChanged(v.uWidth, width);
        ValueCell.updateIfChanged(v.uNCenters, data.nCenters);
        ValueCell.updateIfChanged(v.uNAlpha, data.nAlpha);
        ValueCell.updateIfChanged(v.uNCoeff, data.nCoeff);
        ValueCell.updateIfChanged(v.uMaxCoeffs, data.maxCoeffs);
        ValueCell.update(v.tCenters, { width: data.nCenters, height: 1, array: data.centers });
        ValueCell.update(v.tInfo, { width: data.nCenters, height: 1, array: data.info });
        ValueCell.update(v.tCoeff, { width: data.nCoeff, height: 1, array: data.coeff });
        ValueCell.update(v.tAlpha, { width: data.nAlpha, height: 1, array: data.alpha });
        ValueCell.updateIfChanged(v.uLittleEndian, isLittleEndian());
        ValueCell.updateIfChanged(v.uDensity, false);
        ValueCell.updateIfChanged(v.uOccupancy, 0);
        ValueCell.updateIfChanged(v.tCumulativeSum, ctx.namedTextures[AlphaOrbitalsTex1]);
        ctx.namedComputeRenderables[AlphaOrbitalsName].update();
    }
    else {
        ctx.namedComputeRenderables[AlphaOrbitalsName] = createAlphaOrbitalsRenderable(ctx, grid, orbital);
    }
    return ctx.namedComputeRenderables[AlphaOrbitalsName];
}
export function gpuComputeAlphaOrbitalsGridValues(webgl, grid, orbital) {
    var _a = grid.dimensions, nx = _a[0], ny = _a[1], nz = _a[2];
    var renderable = getAlphaOrbitalsRenderable(webgl, grid, orbital);
    var width = renderable.values.uWidth.ref.value;
    var framebuffer = webgl.namedFramebuffers[AlphaOrbitalsName];
    webgl.namedTextures[AlphaOrbitalsTex0].define(width, width);
    webgl.namedTextures[AlphaOrbitalsTex0].attachFramebuffer(framebuffer, 'color0');
    var gl = webgl.gl, state = webgl.state;
    framebuffer.bind();
    gl.viewport(0, 0, width, width);
    gl.scissor(0, 0, width, width);
    state.disable(gl.SCISSOR_TEST);
    state.disable(gl.BLEND);
    state.disable(gl.DEPTH_TEST);
    state.depthMask(false);
    renderable.render();
    var array = new Uint8Array(width * width * 4);
    webgl.readPixels(0, 0, width, width, array);
    return new Float32Array(array.buffer, array.byteOffset, nx * ny * nz);
}
export function canComputeAlphaOrbitalsOnGPU(webgl) {
    return !!(webgl === null || webgl === void 0 ? void 0 : webgl.extensions.textureFloat);
}
export function gpuComputeAlphaOrbitalsDensityGridValues(webgl, grid, orbitals, ctx) {
    return __awaiter(this, void 0, void 0, function () {
        var _a, nx, ny, nz, renderable, width, framebuffer, tex, values, gl, state, nonZero, i, alpha, array;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0: return [4 /*yield*/, ctx.update({ message: 'Initializing...', isIndeterminate: true })];
                case 1:
                    _b.sent();
                    _a = grid.dimensions, nx = _a[0], ny = _a[1], nz = _a[2];
                    renderable = getAlphaOrbitalsRenderable(webgl, grid, orbitals[0]);
                    width = renderable.values.uWidth.ref.value;
                    if (!webgl.namedFramebuffers[AlphaOrbitalsName]) {
                        webgl.namedFramebuffers[AlphaOrbitalsName] = webgl.resources.framebuffer();
                    }
                    framebuffer = webgl.namedFramebuffers[AlphaOrbitalsName];
                    tex = [webgl.namedTextures[AlphaOrbitalsTex0], webgl.namedTextures[AlphaOrbitalsTex1]];
                    tex[0].define(width, width);
                    tex[1].define(width, width);
                    values = renderable.values;
                    gl = webgl.gl, state = webgl.state;
                    gl.viewport(0, 0, width, width);
                    gl.scissor(0, 0, width, width);
                    state.disable(gl.SCISSOR_TEST);
                    state.disable(gl.BLEND);
                    state.disable(gl.DEPTH_TEST);
                    state.depthMask(false);
                    gl.clearColor(0, 0, 0, 0);
                    tex[0].attachFramebuffer(framebuffer, 'color0');
                    gl.clear(gl.COLOR_BUFFER_BIT);
                    tex[1].attachFramebuffer(framebuffer, 'color0');
                    gl.clear(gl.COLOR_BUFFER_BIT);
                    ValueCell.update(values.uDensity, true);
                    nonZero = orbitals.filter(function (o) { return o.occupancy !== 0; });
                    return [4 /*yield*/, ctx.update({ message: 'Computing...', isIndeterminate: false, current: 0, max: nonZero.length })];
                case 2:
                    _b.sent();
                    i = 0;
                    _b.label = 3;
                case 3:
                    if (!(i < nonZero.length)) return [3 /*break*/, 6];
                    alpha = getNormalizedAlpha(grid.params.basis, nonZero[i].alpha, grid.params.sphericalOrder);
                    ValueCell.update(values.uOccupancy, nonZero[i].occupancy);
                    ValueCell.update(values.tCumulativeSum, tex[(i + 1) % 2]);
                    ValueCell.update(values.tAlpha, { width: alpha.length, height: 1, array: alpha });
                    tex[i % 2].attachFramebuffer(framebuffer, 'color0');
                    gl.viewport(0, 0, width, width);
                    gl.scissor(0, 0, width, width);
                    state.disable(gl.SCISSOR_TEST);
                    state.disable(gl.BLEND);
                    state.disable(gl.DEPTH_TEST);
                    state.depthMask(false);
                    renderable.update();
                    renderable.render();
                    if (!(i !== nonZero.length - 1 && ctx.shouldUpdate)) return [3 /*break*/, 5];
                    return [4 /*yield*/, ctx.update({ current: i + 1 })];
                case 4:
                    _b.sent();
                    _b.label = 5;
                case 5:
                    i++;
                    return [3 /*break*/, 3];
                case 6:
                    array = new Uint8Array(width * width * 4);
                    webgl.readPixels(0, 0, width, width, array);
                    return [2 /*return*/, new Float32Array(array.buffer, array.byteOffset, nx * ny * nz)];
            }
        });
    });
}
//# sourceMappingURL=compute.js.map